//
// Created by robin on 2/2/18.
//

#include <QtCore/QTimer>
#include "dns-client/QDnsClient.h"

namespace QDns
{
    namespace Client
    {
        QDnsClient::QDnsClient(const QHostAddress &host, quint16 port)
            : m_host(host)
            , m_port(port)
            , m_queryTimeout(5000)
            , m_state(ClientState::Disconnected)
        {
        }

        const QHostAddress &QDnsClient::getHost() const
        {
            return m_host;
        }

        quint16 QDnsClient::getPort() const
        {
            return m_port;
        }

        quint32 QDnsClient::getQueryTimeout() const
        {
            return m_queryTimeout;
        }

        void QDnsClient::setQueryTimeout(quint32 queryTimeout)
        {
            m_queryTimeout = queryTimeout;
        }

        QDnsClient::ClientState QDnsClient::getState() const
        {
            return m_state;
        }

        void QDnsClient::query(QDns::Base::QDnsPacket packet)
        {
            if (m_state != ClientState::Connected)
            {
                emit onError(OnErrorParams
                                     {
                                             0,
                                             ErrorCode::NotConnected,
                                             QAbstractSocket::UnknownSocketError,
                                             QDns::Base::QDnsPacket(""),
                                             QByteArray(),
                                             packet,
                                             m_host,
                                             m_port
                                     });
                return;
            }
            if (packet.getHeader().getId() == 0)
            {
                auto header = packet.getHeader();
                header.setId((quint16) qrand());
                packet.setHeader(header);
            }
            m_pendingQueries.insert(packet.getHeader().getId(), packet);

            auto timer = new QTimer(this);
            connect(timer, &QTimer::timeout, [this, packet]()
            {
                if (m_pendingQueries.contains(packet.getHeader().getId()))
                {
                    emit onError(OnErrorParams
                                         {
                                                 0,
                                                 ErrorCode::Timeout,
                                                 QAbstractSocket::UnknownSocketError,
                                                 packet,
                                                 QByteArray(),
                                                 QDns::Base::QDnsPacket(""),
                                                 m_host,
                                                 m_port
                                         });
                }
            });
            timer->setSingleShot(true);
            timer->start(m_queryTimeout);

            doQuery(packet);
        }

        void QDnsClient::processData(QDnsClient::ProcessDataParams params)
        {
            QDns::Base::QDnsPacket packet(params.data);
            if (packet.isValid())
            {
                ErrorCode error = ErrorCode::Unknown;
                QDns::Base::QDnsPacket query("");
                if (!m_pendingQueries.contains(packet.getHeader().getId()))
                {
                    error = ErrorCode::UnknownId;
                }
                else
                {
                    query = m_pendingQueries[packet.getHeader().getId()];
                    m_pendingQueries.remove(packet.getHeader().getId());
                }
                if (packet.getHeader().isQuery())
                {
                    error = ErrorCode::NotAnswer;
                }
                if (error == ErrorCode::Unknown)
                {
                    emit onAnswer(OnAnswerParams
                                          {
                                                  0,
                                                  query,
                                                  packet,
                                                  params.senderAddress,
                                                  params.senderPort
                                          });
                }
                else
                {
                    emit onError(OnErrorParams
                                         {
                                                 0,
                                                 error,
                                                 QAbstractSocket::UnknownSocketError,
                                                 query,
                                                 params.data,
                                                 packet,
                                                 params.senderAddress,
                                                 params.senderPort
                                         });
                }
            }
            else
            {
                emit onError(OnErrorParams
                                     {
                                             0,
                                             ErrorCode::InvalidData,
                                             QAbstractSocket::UnknownSocketError,
                                             QDns::Base::QDnsPacket(""),
                                             params.data,
                                             QDns::Base::QDnsPacket(""),
                                             params.senderAddress,
                                             params.senderPort
                                     });
            }
        }

        void QDnsClient::setState(QDnsClient::ClientState state)
        {
            m_state = state;
            emit onStateChanged(OnStateChangedParams
                                        {
                                                0,
                                                m_state
                                        });
        }
    }
}